"""IntexSpa"""
import logging
import time
import asyncio

import typing

from .intex_spa_network_layer import IntexSpaNetworkLayer
from .intex_spa_query import IntexSpaQuery
from .intex_spa_status import IntexSpaStatus

_LOGGER = logging.getLogger(__name__)


class IntexSpa:
    """
    AsyncIO-enabled class to interface with Intex Spa wifi module

    Attributes
    -------
    network : IntexSpaNetworkLayer
        The network layer object for communications with intex spa wifi module
    status : IntexSpaStatus
        The status object of the spa
    last_successful_update_ms : int
        The millisecond timestamp of the last successful update
    is_available : bool
        Is the spa wifi module available
    """

    def __init__(self, address: str = "SPA_DEVICE", port: str = "8990"):
        """
        Initialize IntexSpa class

        Parameters
        ----------
        address : str, default = "SPA_DEVICE"
            The fqdn or IP of the intex spa wifi module
        port : str, default = "8990"
            The TCP service port the intex spa wifi module
        """
        _LOGGER.warning("Initializing IntexSpa instance")
        self.network = IntexSpaNetworkLayer(address, port)
        self.status = IntexSpaStatus()
        self.platforms = {
            "switches": [
                "power",
                "filter",
                "jets",
                "bubbles",
                "sanitizer",
            ],
            "climates": [
                {
                    "unit": "unit",
                    "current_temp": "current_temp",
                    "preset_temp": "preset_temp",
                    "heater": "heater",
                },
            ],
        }
        self.last_successful_update_ms: int = None
        self.is_available: bool = None

    async def _async_handle_intent(
        self, intent: str = "status", expected_state: typing.Union[bool, int] = None
    ) -> IntexSpaStatus:
        """
        Handle any intent by conversing with the spa wifi module

        An intent can be:
        * update: to refresh the status object of the spa
        * *command*: to change the function state or temperature preset on the spa

        A conversation with the spa is made of one or more queries:
        * A command intent always triggers a preliminary status update
        * Command is sent and response is received
        * Retries can happen as needed

        Parameters
        ----------
        intent : str
            The intent to handle (i.e.: "status", "heater", "preset_temp", ...)
        expected_state : bool | int, optional
            The expected state of the function or the temperature preset

        Returns
        -------
        status : IntexSpaStatus
            The status of the spa
        """

        _LOGGER.debug("'%s' intent: Handling new intent...", intent)

        # Trigger a preliminary update status intent if the provided intent is a command
        if intent != "status":
            _LOGGER.debug(
                "'%s' intent: triggering a preliminary 'update' intent", intent
            )
            await self.async_update_status()

        if (
            # the provided intent is an update status
            intent == "status"
            # the provided intent is a command and its expected_state differs from the current state
            or getattr(self.status, intent) != expected_state
        ):
            _LOGGER.debug("'%s' intent: a spa query is needed", intent)

            # Attempt maximum 5 times
            for _ in range(5):
                try:
                    _LOGGER.debug("'%s' intent: new spa query", intent)
                    # Initialize a query to the spa
                    query = IntexSpaQuery(intent, expected_state)

                    # Send the raw request bytes via the network object
                    await self.network.async_send(query.request_bytes)

                    # Receive the raw response bytes via the network object
                    received_bytes = await self.network.async_receive()
                    # Give the raw received_bytes back to the query object to render the new status
                    query.render_response_status(received_bytes)
                    _LOGGER.debug("'%s' intent: new status is rendered", intent)
                    # And update the status object with it
                    self.status = query.response_status

                    # Set availability info
                    self.last_successful_update_ms = int(time.time() * 1000)
                    self.is_available = True
                except (
                    asyncio.TimeoutError,
                    asyncio.IncompleteReadError,
                    AssertionError,
                ):
                    _LOGGER.warning("Exception raised during spa querying")
                    await asyncio.sleep(2)
                    continue
                else:
                    break
            else:  # No retry has succeeded
                # Set unavailability info
                self.status = IntexSpaStatus()
                self.is_available = False

        return self.status

    async def async_update_status(self) -> IntexSpaStatus:
        """Update known status of the spa

        Returns
        -------
        status : IntexSpaStatus
            The updated spa status
        """
        return await self._async_handle_intent("status")

    async def async_set_power(self, expected_state: bool = True) -> IntexSpaStatus:
        """
        Set power function to `expected_state`

        Returns
        -------
        status : IntexSpaStatus
            The updated spa status
        """
        return await self._async_handle_intent("power", expected_state)

    async def async_set_filter(self, expected_state: bool = True) -> IntexSpaStatus:
        """
        Set filter function to `expected_state`

        Returns
        -------
        status : IntexSpaStatus
            The updated spa status
        """
        return await self._async_handle_intent("filter", expected_state)

    async def async_set_heater(self, expected_state: bool = True) -> IntexSpaStatus:
        """
        Set heater function to `expected_state`

        Returns
        -------
        status : IntexSpaStatus
            The updated spa status
        """
        return await self._async_handle_intent("heater", expected_state)

    async def async_set_jets(self, expected_state: bool = True) -> IntexSpaStatus:
        """
        Set jets function to `expected_state`

        Returns
        -------
        status : IntexSpaStatus
            The updated spa status
        """
        return await self._async_handle_intent("jets", expected_state)

    async def async_set_bubbles(self, expected_state: bool = True) -> IntexSpaStatus:
        """
        Set bubbles function to `expected_state`

        Returns
        -------
        status : IntexSpaStatus
            The updated spa status
        """
        return await self._async_handle_intent("bubbles", expected_state)

    async def async_set_sanitizer(self, expected_state: bool = True) -> IntexSpaStatus:
        """
        Set sanitizer function to `expected_state`

        Returns
        -------
        status : IntexSpaStatus
            The updated spa status
        """
        return await self._async_handle_intent("sanitizer", expected_state)

    async def async_set_preset_temp(self, expected_state: int) -> IntexSpaStatus:
        """
        Set preset_temp function to `expected_state`

        Returns
        -------
        status : IntexSpaStatus
            The updated spa status
        """
        return await self._async_handle_intent("preset_temp", expected_state)



"""IntexSpaQuery"""
import time
import json

from .intex_spa_status import IntexSpaStatus

REQUEST: dict = {
    "status": "8888060FEE0F01",
    "power": "8888060F014000",
    "filter": "8888060F010004",
    "heater": "8888060F010010",
    "jets": "8888060F011000",
    "bubbles": "8888060F010400",
    "sanitizer": "8888060F010001",
    "preset_temp": "8888050F0C",
}


def checksum_as_int(data: str) -> int:
    """Return integer checksum for the given data, as expected by Intex Spa protocol"""
    calculated_checksum = 0xFF
    for index in range(0, len(data), 2):
        calculated_checksum = calculated_checksum - (
            int("0x" + data[index : index + 2], 16)
        )
    return calculated_checksum % 0xFF


def checksum_as_str(data: str) -> str:
    """Return string checksum for the given data, as expected by Intex Spa protocol"""
    # Return checksum as a hex string without 0x prefix
    return hex(checksum_as_int(data) % 0xFF)[2:].upper()


class IntexSpaQuery:
    """
    Class to manage one application-layer query with Intex Spa wifi module

    Manages encoding and decoding of one request and its response messages

    Attributes
    -------
    intex_timestamp : str
        The 10th of milliseconds timestamp, as expected by Intex Spa protocol
    request : str
        The request data to send, as expected by Intex Spa protocol
    request_bytes : bytes
        The full message request to send, as expected by Intex Spa protocol, encoded as bytes
    response_status : IntexSpaStatus
        The rendered response data, as int
    """

    def __init__(self, intent: str, preset_temp: int = None):
        """Init."""
        self.intex_timestamp = str(int(time.time() * 10000))

        self.request: str = REQUEST[intent]
        if intent == "preset_temp":
            self.request = self.request + hex(preset_temp)[2:].upper()

        self.response_status: IntexSpaStatus

    @property
    def request_bytes(self) -> bytes:
        """The full message request to send, as expected by Intex Spa protocol, encoded as bytes"""
        request_dict = {
            "data": self.request + checksum_as_str(self.request),
            "sid": self.intex_timestamp,
            "type": 1,
        }
        return json.dumps(request_dict).encode()

    def render_response_status(self, received_bytes: bytes) -> IntexSpaStatus:
        """
        Render response data from `received_bytes` from Intex Spa wifi module

        Parameters
        ----------
        received_bytes : bytes
            The response received from Intex Spa wifi module, as bytes

        Returns
        ----------
        response_status : IntexSpaStatus
            The new status, rendered from the spa response
        """
        response = json.loads(received_bytes.decode())

        # Timestamp correspondance check
        assert response["sid"] == self.intex_timestamp
        assert response["result"] == "ok"
        assert response["type"] == 2

        # Checksum comparison
        checksum_calculated = checksum_as_int(response["data"][:-2])
        checksum_in_response = int("0x" + response["data"][-2:], 16)
        assert checksum_calculated == checksum_in_response

        response_as_int = int("0x" + response["data"], 16)
        self.response_status = IntexSpaStatus(response_as_int)

        return self.response_status



"""IntexSpa"""
import logging
import asyncio

_LOGGER = logging.getLogger(__name__)


class IntexSpaNetworkLayer:
    """
    AsyncIO-enabled class to manage network-layer communications with Intex Spa wifi module

    Attributes
    -------
    address : str
        The fqdn or IP of the intex spa wifi module
    port : str
        The TCP service port the intex spa wifi module
    reader : asyncio.StreamReader
        The StreamReader connected to the spa
    writer : asyncio.StreamWriter
        The StreamWriter connected to the spa
    """

    def __init__(self, address: str, port: str):
        """
        Initialize IntexSpaNetworkLayer class

        Parameters
        ----------
        address : str
            The fqdn or IP of the intex spa wifi module
        port : str
            The TCP service port of the intex spa wifi module
        """
        self.address = address
        self.port = port

        self.reader: asyncio.StreamReader = None
        self.writer: asyncio.StreamWriter = None

    async def _async_connect(self) -> None:
        """Initialize a connection to the spa"""
        _LOGGER.debug(
            "Opening TCP connection with the spa at %s:%s with asyncio",
            self.address,
            self.port,
        )
        self.reader, self.writer = await asyncio.open_connection(
            self.address, self.port
        )
        _LOGGER.info(
            "TCP connection established with the spa",
        )

    async def _async_disconnect(self) -> None:
        """Close the connection to the spa"""
        _LOGGER.debug(
            "Closing TCP connection to the spa with asyncio",
        )
        self.writer.close()
        await self.writer.wait_closed()
        _LOGGER.info(
            "TCP connection closed with the spa",
        )

    async def async_send(self, bytes_to_write: bytes = None) -> None:
        """Send command to the spa"""
        if self.writer is None or self.reader is None:
            _LOGGER.info("Not connected to the spa")
            await self._async_connect()
        elif self.writer.is_closing() or self.reader.at_eof():
            _LOGGER.info("Connection with the spa seems to be broken")
            await self._async_disconnect()
            await self._async_connect()

        _LOGGER.debug("Sending bytes to the spa: %s", bytes_to_write)
        self.writer.write(bytes_to_write)
        await self.writer.drain()

    async def async_receive(self) -> None:
        """Receive response from the spa"""
        response_as_bytes = await self.reader.readline()
        _LOGGER.debug("Receiving bytes from the spa: %s", response_as_bytes)
        return response_as_bytes



"""IntexSpaStatus"""
import logging

_LOGGER = logging.getLogger(__name__)


class IntexSpaStatus:
    """
    Class to represent Intex Spa status

    Attributes
    -------
    _raw_status : int
        The raw integer-encoded status data, as received from the spa
    power : bool
    filter : bool
    heater : bool
    jets : bool
    bubbles : bool
    sanitizer : bool
    unit : str
    current_temp : int
    preset_temp : int
    """

    @property
    def power(self) -> bool:
        """Power state of the spa"""
        return bool((self._raw_status >> 104) & 0b1)

    @property
    def filter(self) -> bool:
        """State of the filter function"""
        return bool((self._raw_status >> 105) & 0b1)

    @property
    def heater(self) -> bool:
        """State of the heater function"""
        return bool((self._raw_status >> 106) & 0b1)

    @property
    def jets(self) -> bool:
        """State of the jets function"""
        return bool((self._raw_status >> 107) & 0b1)

    @property
    def bubbles(self) -> bool:
        """State of the bubbles function"""
        return bool((self._raw_status >> 108) & 0b1)

    @property
    def sanitizer(self) -> bool:
        """State of sanitizer function"""
        return bool((self._raw_status >> 109) & 0b1)

    @property
    def unit(self) -> str:
        """Temperature measurement unit
        *"°C" for Celsius*
        *"°F" for Farenheit*
        """
        if self.preset_temp <= 40:
            return "°C"
        else:
            return "°F"

    @property
    def current_temp(self) -> int:
        """Current temperature of the water, expressed in `unit`"""
        return (self._raw_status >> 88) & 0xFF

    @property
    def preset_temp(self) -> int:
        """Preset temperature of the water, expressed in `unit`"""
        return (self._raw_status >> 24) & 0xFF

    def __init__(self, raw_status: int = None):
        """
        Initialize IntexSpaStatus class

        Parameters
        ----------
        raw_status : int, optional
            The raw response data received from the spa
        """
        if raw_status is not None:
            self.update(raw_status)

    def update(self, raw_status: int):
        """
        Update the raw_status

        Parameters
        ----------
        raw_status : int
            The raw response data received from the spa
        """
        self._raw_status = raw_status
        _LOGGER.debug("Spa status: '%s'", self)

    def as_dict(self) -> dict:
        """
        Return main status attributes only, as dict

        Returns
        -------
        status_attributes : dict
            IntexSpaStatus main status attributes as dict
        """
        return {
            "power": self.power,
            "filter": self.filter,
            "heater": self.heater,
            "jets": self.jets,
            "bubbles": self.bubbles,
            "sanitizer": self.sanitizer,
            "unit": self.unit,
            "current_temp": self.current_temp,
            "preset_temp": self.preset_temp,
        }

    def __repr__(self) -> str:
        """
        Represent IntexSpaStatus main status attributes
        """
        return repr(self.as_dict())



"""Setup configuration."""
from setuptools import setup, find_packages


with open("README.md", "r") as fh:
    README = fh.read()
setup(
    name="intex-spa",
    version="0.3.0",
    author="Mathieu Payrol",
    author_email="intex-spa@payrol.fr",
    description="Intex Spa wifi client",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/mathieu-mp/intex-spa",
    packages=find_packages(),
    install_requires=["asyncio"],
    classifiers=(
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ),
)
